#extension GL_ARB_texture_rectangle : enable

uniform sampler2DRect texEntry;
uniform sampler2DRect texExit;
uniform sampler3D sliceSampler;
uniform sampler1D transferSampler;
uniform int mode;
uniform float dz;
uniform bool light;
uniform bool compMode;

varying mat4 rotMat;

const int maxrange = 347;
const vec3 lightColor = vec3(1.0, 1.0, 1.0);
const vec4 lightVector = vec4(-1.0, -1.0, -1.0, 1.0);

// gradient shading
vec3 shaded(vec3 normal, vec3 view, vec3 light, vec3 color)
{
	float aI = 0.2;
	float dI = 0.8;
	float sI = 1.0;
	float sE = 100.0;
	
	float dF = max(dot(normal, light), 0.0);
	float sF = pow(max(dot(normal, normalize(light + view)), 0.0), sE);
	
	vec3 result;
	result.x = color.x * aI
		+ (color.x * lightColor.x * dF * dI)
		+ lightColor.x * (sF * sI);
	result.y = color.y * aI
		+ (color.y * lightColor.y * dF * dI)
		+ lightColor.y * (sF * sI);
	result.z = color.z * aI
		+ (color.z * lightColor.z * dF * dI)
		+ lightColor.z * (sF * sI);
	
	return result;
}

// f2b compositing: mode = 0
vec4 compositingF2B(vec3 startPos, vec3 diff, int maxiter)
{
	vec3 pos = startPos;
	vec4 result = vec4(0.0, 0.0, 0.0, 0.0);
	int i = 0;
	
	for(i = 0; i < maxrange; i++)
	{
		vec4 intensity = texture3D(sliceSampler, pos);
		vec4 transf = texture1D(transferSampler, intensity.x);
		

		if(light)
		{
			float delta = 0.01;
			
			vec3 sample1 = vec3(
				texture3D(sliceSampler, pos - vec3(delta, 0.0, 0.0)).x,
				texture3D(sliceSampler, pos - vec3(0.0, delta, 0.0)).x,
				texture3D(sliceSampler, pos - vec3(0.0, 0.0, delta)).x);
			vec3 sample2 = vec3(
				texture3D(sliceSampler, pos + vec3(delta, 0.0, 0.0)).x,
				texture3D(sliceSampler, pos + vec3(0.0, delta, 0.0)).x,
				texture3D(sliceSampler, pos + vec3(0.0, 0.0, delta)).x);
			
			vec4 normal = vec4(normalize(sample2 - sample1), 1.0);
			
			result.xyz += (1.0 - result.w) * transf.w
				* shaded(
					normal.xyz,
					diff,
					normalize((lightVector * rotMat).xyz),
					transf.xyz);
		}
		else
			result.xyz += (1.0 - result.w) * transf.w * transf.xyz;
			
		result.w += (1.0 - result.w) * transf.w;
		
		pos += diff;
		
		if((i >= maxiter) || (result.w >= 1.0))
			break;
	}
	
	return result;
}

// b2f compositing: mode = 0
vec4 compositingB2F(vec3 startPos, vec3 diff, int maxiter)
{
	vec3 pos = startPos;
	vec4 result = vec4(0.0, 0.0, 0.0, 0.0);
	int i = 0;
	
	for(i = 0; i < maxrange; i++)
	{
		vec4 intensity = texture3D(sliceSampler, pos);
		vec4 transf = texture1D(transferSampler, intensity.x);
		

		if(light)
		{
			float delta = 0.01;
			
			vec3 sample1 = vec3(
				texture3D(sliceSampler, pos - vec3(delta, 0.0, 0.0)).x,
				texture3D(sliceSampler, pos - vec3(0.0, delta, 0.0)).x,
				texture3D(sliceSampler, pos - vec3(0.0, 0.0, delta)).x);
			vec3 sample2 = vec3(
				texture3D(sliceSampler, pos + vec3(delta, 0.0, 0.0)).x,
				texture3D(sliceSampler, pos + vec3(0.0, delta, 0.0)).x,
				texture3D(sliceSampler, pos + vec3(0.0, 0.0, delta)).x);
			
			vec4 normal = vec4(normalize(sample2 - sample1), 1.0);
			
			result = result * (1.0 - transf.w)
				+ transf.w * vec4(shaded(
					normal.xyz,
					diff,
					normalize((lightVector * rotMat).xyz),
					transf.xyz), 1.0);
		}
		else
			result = result * (1.0 - transf.w)
				+ transf.w * vec4(transf.xyz, 1.0);
		
		pos -= diff;
		
		if((i >= maxiter))
			break;
	}
	
	return result;
}

// MIP: mode = 1
vec4 mip(vec3 startPos, vec3 diff, int maxiter)
{
	vec3 pos = startPos;
	float result = 0.0;

	for(int i = 0; i < maxrange; i++)
	{
		float intensity = (texture3D(sliceSampler, pos)).x;
		
		if(intensity > result)
			result = intensity;
		
		pos += diff;
		
		if(i >= maxiter)
			break;
	}
	
	return texture1D(transferSampler, result);
}

// mode = 2
vec4 averaging(vec3 startPos, vec3 diff, int maxiter)
{
	vec3 pos = startPos;
	vec4 result = vec4(0.0, 0.0, 0.0, 0.0);
	int i = 0;
	
	for(i = 0; i < maxrange; i++)
	{
		vec4 intensity = texture3D(sliceSampler, pos);
		vec4 transf = texture1D(transferSampler, intensity.x);
		
		result.xyz += transf.w * transf.xyz;
		result.w += transf.w;
		
		pos += diff;
		
		if(i >= maxiter)
			break;
	}
	
	return result / float(i);
}

void main()
{	
	vec4 startPos = texture2DRect(texEntry, vec2(gl_FragCoord));
	
	if(startPos.w == 0.0)
		discard;
	
	vec4 endPos = texture2DRect(texExit, vec2(gl_FragCoord));
	
	float dist = distance(startPos, endPos) / dz;
	int maxiter = int(floor(dist));
	vec3 diff = (endPos.xyz - startPos.xyz) / dist;
	
	if(mode == 0 && compMode)
		gl_FragColor = compositingF2B(startPos.xyz, diff, maxiter);
	else if(mode == 0 && !compMode)
		gl_FragColor = compositingB2F(endPos.xyz, diff, maxiter);
	else if(mode == 1)
		gl_FragColor = mip(startPos.xyz, diff, maxiter);
	else if(mode == 2)
		gl_FragColor = averaging(startPos.xyz, diff, maxiter);
		
	//gl_FragColor = startPos;
}